from yta_multimedia.video.edition.effect.m_effect import MEffect as Effect
from yta_multimedia.video.parser import VideoParser
from moviepy import Clip
from moviepy.video.fx import Scroll


class ScrollEffect(Effect):
    """
    This effect will make the clip be scrolled like if a zoomed
    region was surfing through the clip.
    """

    def apply(self, video: Clip, width = None, height = None, x_speed = None, y_speed = None, x_start = None, y_start = None) -> Clip:
        # TODO: This is not working properly yet
        #PythonValidator.validate_method_params(BlinkEffect.apply, locals(), ['video'])
        video = VideoParser.to_moviepy(video)

        width = width if width else video.w / 2
        height = height if height else video.h / 2
        # TODO: Make these params below dynamic
        x_speed = x_speed if x_speed else 20
        y_speed = y_speed if y_speed else 20
        x_start = x_start if x_start else 0
        y_start = y_start if y_start else 0

        return Scroll(width, height, x_speed, y_speed, x_start, y_start)


from yta_multimedia.video.edition.effect.m_effect import MEffect as Effect
from yta_multimedia.video.parser import VideoParser
from yta_multimedia.video.edition.effect.fade_out_effect import FadeOutEffect
from yta_multimedia.video.edition.effect.fade_in_effect import FadeInEffect
from yta_general_utils.color import Color, ColorString
from moviepy import Clip, concatenate_videoclips
from typing import Union


class BlinkEffect(Effect):
    """
    This method makes the provided video blink, that is a composition of
    a FadeOut and a FadeIn consecutively to build this effect. The duration
    will be the whole clip duration. The FadeIn will last the half of the
    clip duration and the FadeOut the other half.

    The 'color' parameter is the color you want for the blink effect as the
    background color. The default value is black ([0, 0, 0]).
    """
    def apply(self, video: Clip, color: Union[list, tuple, str, Color] = None) -> Clip:
        # TODO: This is not working properly yet
        #PythonValidator.validate_method_params(BlinkEffect.apply, locals(), ['video'])
        video = VideoParser.to_moviepy(video)

        color = Color.parse(color).as_rgb_array() if color is not None else Color.parse(ColorString.BLACK).as_rgb_array()

        half_duration = video.duration / 2
        video = concatenate_videoclips([
            FadeOutEffect().apply(video.with_subclip(0, half_duration), half_duration, color),
            FadeInEffect().apply(video.with_subclip(half_duration, video.duration), half_duration, color)
        ])

        return video


from yta_multimedia.video.generation.manim.constants import HALF_SCENE_HEIGHT, HALF_SCENE_WIDTH
from yta_multimedia.video.edition.effect.moviepy.position.objects.coordinate import Coordinate
from yta_general_utils.random import randrangefloat
from yta_general_utils.programming.enum import YTAEnum as Enum
from yta_general_utils.coordinate.coordinate import NORMALIZATION_MAX_VALUE
from yta_general_utils.programming.parameter_validator import NumberValidator, PythonValidator
from random import choice as randchoice


class Position(Enum):
    """
    Enum class that represents a position within an specific
    scene. This is used to position a video or an image in that
    scene in an specific position defined by itself. It is
    useful with Manim and Moviepy video positioning and has
    been prepared to work with those engines.

    This position is always considered in a scene of 1920x1080.

    This position is a static position which means that it can
    be calculated by itself, without the need of any element
    size or similar. The coordinate that represent any of this
    """
    CENTER = 'center'
    """
    Just on the center of the scene
    """
    IN_EDGE_TOP = 'in_edge_top'
    """
    Just on the top edge of the scene.
    """
    IN_EDGE_BOTTOM = 'in_edge_bottom'
    IN_EDGE_RIGHT = 'in_edge_right'
    IN_EDGE_LEFT = 'in_edge_left'

    IN_EDGE_TOP_LEFT = 'in_edge_top_left'
    """
    Just on the upper left edge corner of the scene.
    """
    IN_EDGE_TOP_RIGHT = 'in_edge_top_right'
    IN_EDGE_BOTTOM_RIGHT = 'in_edge_bottom_right'
    IN_EDGE_BOTTOM_LEFT = 'in_edge_bottom_left'

    HALF_LEFT = 'half_left'
    """
    Between the left edge and the center of the scene.
    """
    HALF_TOP_LEFT = 'half_top_left'
    HALF_TOP = 'half_top'
    HALF_TOP_RIGHT = 'half_top_right'
    HALF_RIGHT = 'half_right'
    HALF_BOTTOM_RIGHT = 'half_bottom_right'
    HALF_BOTTOM = 'half_bottom'
    HALF_BOTTOM_LEFT = 'half_bottom_left'

    OUT_LEFT = 'out_left'
    """
    Imagine our scene surrounded by another 8 scenes.
    This will be at the center that surrounding scene
    that is on the left of this one.
    """
    OUT_TOP = 'out_top'
    """
    Imagine our scene surrounded by another 8 scenes.
    This will be at the center that surrounding scene
    that is on the top of this one.
    """
    OUT_RIGHT = 'out_right'
    """
    Imagine our scene surrounded by another 8 scenes.
    This will be at the center that surrounding scene
    that is on the right of this one.
    """
    OUT_BOTTOM = 'out_bottom'
    """
    Imagine our scene surrounded by another 8 scenes.
    This will be at the center that surrounding scene
    that is on the bottom of this one.
    """
    OUT_TOP_LEFT = 'out_top_left'
    """
    Imagine our scene surrounded by another 8 scenes.
    This will be at the center that surrounding scene
    that is on the top and left of this one.
    """
    OUT_TOP_RIGHT = 'out_top_right'
    """
    Imagine our scene surrounded by another 8 scenes.
    This will be at the center that surrounding scene
    that is on the top and right of this one.
    """
    OUT_BOTTOM_LEFT = 'out_bottom_left'
    """
    Imagine our scene surrounded by another 8 scenes.
    This will be at the center that surrounding scene
    that is on the bottom and left of this one.
    """
    OUT_BOTTOM_RIGHT = 'out_bottom_right'
    """
    Imagine our scene surrounded by another 8 scenes.
    This will be at the center that surrounding scene
    that is on the bottom and right of this one.
    """
    RANDOM_INSIDE = 'random_inside'
    RANDOM_OUTSIDE = 'random_outside'

    @staticmethod
    def get_outside_elements_list():
        """
        Return all the positions that are outside of 
        the screen as a list.
        """
        return [
            Position.OUT_BOTTOM,
            Position.OUT_BOTTOM_LEFT,
            Position.OUT_BOTTOM_RIGHT,
            Position.OUT_RIGHT,
            Position.OUT_LEFT,
            Position.OUT_TOP,
            Position.OUT_TOP_LEFT,
            Position.OUT_TOP_RIGHT
        ]
    
    @staticmethod
    def get_inside_elements_list():
        """
        Return all the positions that are inside of the 
        screen as a list.
        """
        return list(set(Position.get_all()) - set(Position.get_outside_elements_list()) - set([Position.RANDOM_INSIDE]) - set([Position.RANDOM_OUTSIDE]))

    @staticmethod
    def get_random_outside() -> 'Position':
        """
        Return a random position which is outside of the
        screen.
        """
        return randchoice(Position.get_outside_elements_list())
    
    @staticmethod
    def get_random_inside() -> 'Position':
        """
        Return a random position which is inside of the
        screen.
        """
        return randchoice(Position.get_inside_elements_list())

    def get_moviepy_center_tuple(self, scene_size: tuple = (1920, 1080)):
        """
        Get the position tuple (x, y) of this position according
        to the scene defined by the provided 'scene_size'. That
        position belongs to a coordinate in which you can place
        the center of any element.

        This works considering the upper left corner as the (0, 0)
        coordinate.
        """
        if scene_size is None:
            scene_size = (1920, 1080)
        else:
            validate_size(scene_size, 'scene_size')

        # Alias to simplify :)
        w, h = scene_size

        if self == Position.RANDOM_INSIDE:
            return Position.get_random_inside().get_moviepy_center_tuple()
        elif self == Position.RANDOM_OUTSIDE:
            return Position.get_random_outside().get_moviepy_center_tuple()

        elif self == Position.CENTER:
            position = (w / 2, h / 2)

        elif self == Position.IN_EDGE_TOP:
            position = (w / 2, 0)
        elif self == Position.IN_EDGE_BOTTOM:
            position = (w / 2, h)
        elif self == Position.IN_EDGE_LEFT:
            position = (0, h / 2)
        elif self == Position.IN_EDGE_RIGHT:
            position = (w, h / 2)
        
        elif self == Position.IN_EDGE_TOP_LEFT:
            position = (0, 0)
        elif self == Position.IN_EDGE_TOP_RIGHT:
            position = (w, 0)
        elif self == Position.IN_EDGE_BOTTOM_RIGHT:
            position = (w, h)
        elif self == Position.IN_EDGE_BOTTOM_LEFT:
            position = (0, h)
        
        elif self == Position.HALF_LEFT:
            position = (w / 4, h / 2)
        elif self == Position.HALF_TOP_LEFT:
            position = (w / 4, h / 4)
        elif self == Position.HALF_TOP:
            position = (w / 2, h / 4)
        elif self == Position.HALF_TOP_RIGHT:
            position = (3 * w / 4, h / 4)
        elif self == Position.HALF_RIGHT:
            position = (3 * w / 4, h / 2)
        elif self == Position.HALF_BOTTOM_RIGHT:
            position = (3 * w / 4, 3 * h / 4)
        elif self == Position.HALF_BOTTOM:
            position = (w / 2, 3 * h / 4)
        elif self == Position.HALF_BOTTOM_LEFT:
            position = (w / 4, 3 * h / 4)

        elif self == Position.OUT_LEFT:
            position = (-w, h / 2)
        elif self == Position.OUT_TOP:
            position = (w / 2, -h / 2)
        elif self == Position.OUT_RIGHT:
            position = (w + w / 2, h / 2)
        elif self == Position.OUT_BOTTOM:
            position = (w / 2, h + h / 2)
        elif self == Position.OUT_TOP_LEFT:
            position = (-w, -h / 2)
        elif self == Position.OUT_TOP_RIGHT:
            position = (w + w / 2, -h / 2)
        elif self == Position.OUT_BOTTOM_LEFT:
            position = (-w, h + h / 2)
        elif self == Position.OUT_BOTTOM_RIGHT:
            position = (w + w / 2, h + h / 2)

        return position
    
    def get_moviepy_upper_left_corner_tuple(self, video_size: tuple, scene_size: tuple = (1920, 1080)):
        """
        Get the position tuple (x, y) of this position according
        to the scene defined by the provided 'scene_size'. That 
        position belongs to a coordinate in which we need to put
        the moviepy video upper left corner to make its center be
        placed in this position.

        This method is useful to position a video in this desired
        position as it returns the value that can be directly used
        in 'with_position' method.

        This works considering the upper left corner as the (0, 0)
        coordinate.
        """
        validate_size(video_size, 'video_size')
        
        # Obtain the position in which the center of the video
        # must be placed
        center_position = self.get_moviepy_center_tuple(scene_size)
        # TODO: Maybe force integer value (?)

        # Obtain the position in which we should place the video
        # (the upper left corner) to make the center of the video
        # be on that 'position_in_scene' position
        upper_left_corner = (center_position[0] - video_size[0] / 2, center_position[1] - video_size[1] / 2)

        return upper_left_corner
    
    def get_manim_position_center(self):
        """
        Get the position tuple (x, y, z) of this position according
        to the manim scene. That position belongs to a coordinate
        in which we need to put the manim video center to make its
        center be placed in this position.

        The 'z' value is always 0 as this is for a 2D scene.

        This works considering the upper left corner as the 
        (-HALF_SCENE_WIDTH, -HALF_SCENE_HEIGHT) coordinate.
        """
        # Alias to simplify :)
        hw, hh = HALF_SCENE_WIDTH, HALF_SCENE_WIDTH

        if self == Position.RANDOM_INSIDE:
            return Position.get_random_inside().get_manim_position_center()
        elif self == Position.RANDOM_OUTSIDE:
            return Position.get_random_outside().get_manim_position_center()

        elif self == Position.CENTER:
            position = 0, 0, 0

        elif self == Position.IN_EDGE_TOP:
            position = 0, hh, 0
        elif self == Position.IN_EDGE_BOTTOM:
            position = 0, -hh, 0
        elif self == Position.IN_EDGE_LEFT:
            position = -hw, 0, 0
        elif self == Position.IN_EDGE_RIGHT:
            position = hw, 0, 0

        elif self == Position.IN_EDGE_TOP_LEFT:
            position = -hw, hh, 0
        elif self == Position.IN_EDGE_TOP_RIGHT:
            position = hw, hh, 0
        elif self == Position.IN_EDGE_BOTTOM_RIGHT:
            position = hw, -hh, 0
        elif self == Position.IN_EDGE_BOTTOM_LEFT:
            position = -hw, -hh, 0
        
        elif self == Position.HALF_LEFT:
            position = -hw / 2, 0, 0
        elif self == Position.HALF_TOP_LEFT:
            position = -hw / 2, hh / 2, 0
        elif self == Position.HALF_TOP:
            position = 0, hh / 2, 0
        elif self == Position.HALF_TOP_RIGHT:
            position = hw / 2, hh / 2, 0
        elif self == Position.HALF_RIGHT:
            position = hw / 2, 0, 0
        elif self == Position.HALF_BOTTOM_RIGHT:
            position = hw / 2, -hh / 2, 0
        elif self == Position.HALF_BOTTOM:
            position = 0, -hh / 2, 0
        elif self == Position.HALF_BOTTOM_LEFT:
            position = -hw / 2, -hh / 2, 0

        elif self == Position.OUT_LEFT:
            position = 2 * -hw, 0, 0
        elif self == Position.OUT_TOP:
            position = 0, 2 * hh, 0
        elif self == Position.OUT_RIGHT:
            position = 2 * hw, 0, 0
        elif self == Position.OUT_BOTTOM:
            position = 0, 2 * -hh, 0
        elif self == Position.OUT_TOP_LEFT:
            position = 2 * -hw, 2 * hh
        elif self == Position.OUT_TOP_RIGHT:
            position = 2 * hw, 2 * hh
        elif self == Position.OUT_BOTTOM_LEFT:
            position = 2 * -hw, 2 * -hh
        elif self == Position.OUT_BOTTOM_RIGHT:
            position = 2 * hw, 2 * -hh

        return position

    @staticmethod
    def to_coordinate(position: 'Position'):
        """
        Turn the provided 'position' (that must be a tuple of
        2 elements) a coordinate.
        """
        position = Position.to_enum(position)

        position = position.get_moviepy_center_tuple()

        return Coordinate(position[0], position[1])
    
    # We don't need any upper left for manim because it
    # uses the center of the elements to position them
    
class DependantPosition(Enum):
    """
    Enum class that represents different positions within an
    scene of 1920x1080 that need a calculation because it
    depends on the size of the element we are trying to 
    position.
    """
    # TODO: Avoid the ones in Position that should not
    # be replaced
    TOP = 'top'
    TOP_RIGHT = 'top_right'
    RIGHT = 'right'
    BOTTOM_RIGHT = 'bottom_right'
    BOTTOM = 'bottom'
    BOTTOM_LEFT = 'bottom_left'
    LEFT = 'left'
    TOP_LEFT = 'top_left'

    OUT_TOP = 'out_top'
    OUT_TOP_RIGHT = 'out_top_right'
    OUT_RIGHT = 'out_right'
    OUT_BOTTOM_RIGHT = 'out_bottom_right'
    OUT_BOTTOM = 'out_bottom'
    OUT_BOTTOM_LEFT = 'out_bottom_left'
    OUT_LEFT = 'out_left'
    OUT_TOP_LEFT = 'out_top_left'

    QUADRANT_1_TOP_RIGHT_CORNER = 'quadrant_1_top_right_corner'
    QUADRANT_1_BOTTOM_RIGHT_CORNER = 'quadrant_1_bottom_right_corner'
    QUADRANT_1_BOTTOM_LEFT_CORNER = 'quadrant_1_bottom_left_corner'
    QUADRANT_2_TOP_LEFT_CORNER = 'quadrant_2_top_left_corner'
    QUADRANT_2_BOTTOM_RIGHT_CORNER = 'quadrant_2_bottom_right_corner'
    QUADRANT_2_BOTTOM_LEFT_CORNER = 'quadrant_2_bottom_left_corner'
    QUADRANT_3_TOP_RIGHT_CORNER = 'quadrant_3_top_right_corner'
    QUADRANT_3_TOP_LEFT_CORNER = 'quadrant_3_top_left_corner'
    QUADRANT_3_BOTTOM_LEFT_CORNER = 'quadrant_3_bottom_left_corner'
    QUADRANT_4_TOP_RIGHT_CORNER = 'quadrant_4_top_right_corner'
    QUADRANT_4_TOP_LEFT_CORNER = 'quadrant_4_top_left_corner'
    QUADRANT_4_BOTTOM_RIGHT_CORNER = 'quadrant_4_bottom_right_corner'

    RANDOM_INSIDE = 'random_inside'
    RANDOM_OUTSIDE = 'random_outside'

    @staticmethod
    def _get_outside_items_as_list():
        """
        Get a list of the DependantPosition items that are placed
        outside of the screen limits.
        """
        return [
            DependantPosition.OUT_TOP_LEFT,
            DependantPosition.OUT_TOP,
            DependantPosition.OUT_RIGHT,
            DependantPosition.OUT_BOTTOM_RIGHT,
            DependantPosition.OUT_BOTTOM,
            DependantPosition.OUT_BOTTOM_LEFT,
            DependantPosition.OUT_LEFT
        ]
    
    @staticmethod
    def get_random_outside():
        """
        Get one random DependantPosition element that is placed
        outside of the screen limits.
        """
        return randchoice(DependantPosition._get_outside_items_as_list())

    @staticmethod
    def _get_inside_items_as_list():
        """
        Get a list of the DependantPosition items that are placed
        inside of the screen limits.
        """
        return list(set(DependantPosition.get_all()) - set(DependantPosition._get_outside_items_as_list()) - set([DependantPosition.RANDOM_INSIDE]) - set([DependantPosition.RANDOM_OUTSIDE]))

    @staticmethod
    def get_random_inside():
        """
        Get one random DependantPosition element that is placed
        inside of the screen limits.
        """
        return randchoice(DependantPosition._get_inside_items_as_list())

    def get_moviepy_position_center(self, video_size: tuple = (1920, 1080), scene_size: tuple = (1920, 1080)):
        """
        Get the position tuple (x, y) of this position according
        to the scene defined by the provided 'scene_size'. That 
        position belongs to a coordinate in which we need to put
        the moviepy video center to be placed in the desired
        position.

        This method is useful cannot be used directly to position
        a video as it is returning the coordinate in which the
        center must be placed and moviepy uses the upper left
        corner to be positioned, so consider using the method
        'get_moviepy_position_upper_left_corner' to obtain it.

        This works considering the upper left corner of the scene
        as the (0, 0) coordinate.
        """
        if video_size is None:
            video_size = (1920, 1080)
        else:
            validate_size(video_size, 'video_size')

        if scene_size is None:
            scene_size = (1920, 1080)
        else:
            validate_size(scene_size, 'scene_size')
            
        # Alias to simplify :)
        sw, sh = scene_size
        vw, vh = video_size

        if self == DependantPosition.TOP:
            position = sw / 2, -vh / 2
        elif self == DependantPosition.RIGHT:
            position = sw - vw / 2, sh / 2
        elif self == DependantPosition.BOTTOM:
            position = sw / 2, sh - vh
        elif self == DependantPosition.LEFT:
            position = vw / 2, sh / 2

        elif self == DependantPosition.TOP_RIGHT:
            position = sw - vw / 2, vh / 2
        elif self == DependantPosition.BOTTOM_RIGHT:
            position = sw - vw / 2, sh - vh / 2
        elif self == DependantPosition.BOTTOM_LEFT:
            position = vw / 2, sh - vh / 2
        elif self == DependantPosition.TOP_LEFT:
            position = vw / 2, vh / 2

        elif self == DependantPosition.OUT_TOP:
            position = sw / 2, -vh / 2
        elif self == DependantPosition.OUT_TOP_RIGHT:
            position = sw + vw / 2, -vh / 2
        elif self == DependantPosition.OUT_RIGHT:
            position = sw + vw / 2, sh / 2
        elif self == DependantPosition.OUT_BOTTOM_RIGHT:
            position = sw + vw / 2, sh + vh / 2
        elif self == DependantPosition.OUT_BOTTOM:
            position = sw / 2, sh + vh / 2
        elif self == DependantPosition.OUT_BOTTOM_LEFT:
            position = -vw / 2, sh + vh / 2
        elif self == DependantPosition.OUT_LEFT:
            position = -vw / 2, sh / 2
        elif self == DependantPosition.OUT_TOP_LEFT:
            position = -vw / 2, -vh / 2

        elif self == DependantPosition.QUADRANT_1_TOP_RIGHT_CORNER:
            position = sw / 2 - vw / 2, vh / 2
        elif self == DependantPosition.QUADRANT_1_BOTTOM_RIGHT_CORNER:
            position = sw / 2 - vw / 2, sh / 2 - vh / 2
        elif self == DependantPosition.QUADRANT_1_BOTTOM_LEFT_CORNER:
            position = vw / 2, sh / 2 - vh / 2
        elif self == DependantPosition.QUADRANT_2_TOP_LEFT_CORNER:
            position = sw / 2 + vw / 2, vh / 2
        elif self == DependantPosition.QUADRANT_2_BOTTOM_RIGHT_CORNER:
            position = sw - vw / 2, sh / 2 - vh / 2
        elif self == DependantPosition.QUADRANT_2_BOTTOM_LEFT_CORNER:
            position = sw / 2 + vw / 2, sh / 2 - vh / 2
        elif self == DependantPosition.QUADRANT_3_TOP_RIGHT_CORNER:
            position = sw - vw / 2, sh / 2 + vh / 2
        elif self == DependantPosition.QUADRANT_3_TOP_LEFT_CORNER:
            position = sw / 2 + vw / 2, sh / 2 + vh / 2
        elif self == DependantPosition.QUADRANT_3_BOTTOM_LEFT_CORNER:
            position = sw / 2 + vw / 2, sh - vh / 2
        elif self == DependantPosition.QUADRANT_4_TOP_RIGHT_CORNER:
            position = sw / 2 - vw / 2, sh / 2 + vh / 2
        elif self == DependantPosition.QUADRANT_4_TOP_LEFT_CORNER:
            position = vw / 2, sh / 2 + vh / 2
        elif self == DependantPosition.QUADRANT_4_BOTTOM_RIGHT_CORNER:
            position = sw / 2 - vw / 2, sh - vh / 2

        return position

    def get_moviepy_position_upper_left_corner(self, video_size: tuple, scene_size: tuple = (1920, 1080)):
        """
        Get the position tuple (x, y) of this position according
        to the scene defined by the provided 'scene_size'. That 
        position belongs to a coordinate in which we need to put
        the moviepy video upper left corner to make its center be
        placed in this position.

        This method is useful to position a video in this desired
        position as it returns the value that can be directly used
        in 'with_position' method.

        This works considering the upper left corner as the (0, 0)
        coordinate.
        """
        # Obtain the position in which the center of the video
        # must be placed
        center_position = self.get_moviepy_position_center(video_size, scene_size)
        # TODO: Maybe force integer value (?)

        # Obtain the position in which we should place the video
        # (the upper left corner) to make the center of the video
        # be on that previously obtained 'center_position' position
        upper_left_corner = (center_position[0] - video_size[0] / 2, center_position[1] - video_size[1] / 2)

        return upper_left_corner

    def get_manim_position_center(self, video_size: tuple):
        if video_size is None:
            raise Exception('No "video_size" provided.')
        
        validate_size(video_size, 'video_size')
        
        # Alias to simplify :)
        hsw, hsh = HALF_SCENE_WIDTH, HALF_SCENE_HEIGHT
        vw, vh = video_size

        # TODO: Process
        if self == DependantPosition.TOP:
            position = 0, hsh - vh / 2, 0
        elif self == DependantPosition.TOP_RIGHT:
            position = hsw - vw / 2, hsh - vh / 2, 0
        elif self == DependantPosition.RIGHT:
            position = hsw - vw / 2, 0, 0
        elif self == DependantPosition.BOTTOM_RIGHT:
            position = hsw - vw / 2, -hsh + vh / 2, 0
        elif self == DependantPosition.BOTTOM:
            position = 0, -hsh + vh / 2, 0
        elif self == DependantPosition.BOTTOM_LEFT:
            position = -hsw + vw / 2, -hsh + vh / 2, 0
        elif self == DependantPosition.LEFT:
            position = -hsw + vw / 2, 0, 0
        elif self == DependantPosition.TOP_LEFT:
            position = -hsw + vw / 2, hsh - vh / 2, 0
        elif self == DependantPosition.OUT_TOP_LEFT:
            position = -hsw - vw / 2, hsh + vh / 2, 0
        elif self == DependantPosition.OUT_TOP:
            position = 0, hsh + vh / 2, 0
        elif self == DependantPosition.OUT_TOP_RIGHT:
            position = hsw + vw / 2, hsh + vh / 2, 0
        elif self == DependantPosition.OUT_RIGHT:
            return hsw + vw / 2, 0, 0
        elif self == DependantPosition.OUT_BOTTOM_RIGHT:
            position = hsw + vw / 2, -hsh - vh / 2, 0
        elif self == DependantPosition.OUT_BOTTOM:
            position = 0, -hsh - vh / 2, 0
        elif self == DependantPosition.OUT_BOTTOM_LEFT:
            position = -hsw - vw / 2, -hsh - vh / 2, 0
        elif self == DependantPosition.OUT_LEFT:
            position = -hsw - vh / 2, 0, 0
        elif self == DependantPosition.QUADRANT_1_TOP_RIGHT_CORNER:
            position = -vw / 2, hsh - vh / 2, 0
        elif self == DependantPosition.QUADRANT_1_BOTTOM_RIGHT_CORNER:
            position = -vw / 2, vh / 2, 0
        elif self == DependantPosition.QUADRANT_1_BOTTOM_LEFT_CORNER:
            position = -hsw + vw / 2, vh / 2, 0
        elif self == DependantPosition.QUADRANT_2_TOP_LEFT_CORNER:
            position = vw / 2, hsh - vh / 2, 0
        elif self == DependantPosition.QUADRANT_2_BOTTOM_RIGHT_CORNER:
            position = hsw - vw / 2, vh / 2, 0
        elif self == DependantPosition.QUADRANT_2_BOTTOM_LEFT_CORNER:
            position = vw / 2, vh / 2, 0
        elif self == DependantPosition.QUADRANT_3_TOP_LEFT_CORNER:
            position = vw / 2, -vh / 2, 0
        elif self == DependantPosition.QUADRANT_3_TOP_RIGHT_CORNER:
            position = hsw - vw / 2, -vh / 2, 0
        elif self == DependantPosition.QUADRANT_3_BOTTOM_LEFT_CORNER:
            position = vw / 2, -hsh + vh / 2, 0
        elif self == DependantPosition.QUADRANT_4_TOP_LEFT_CORNER:
            position = -hsw + vw / 2, -vh / 2, 0
        elif self == DependantPosition.QUADRANT_4_TOP_RIGHT_CORNER:
            position = -vw / 2, -vh / 2, 0
        elif self == DependantPosition.QUADRANT_4_BOTTOM_RIGHT_CORNER:
            position = -vw / 2, -hsh + vh / 2, 0
        elif self == DependantPosition.RANDOM_INSIDE:
            position = DependantPosition.get_random_inside().get_manim_position_center((vw, vh))
        elif self == DependantPosition.RANDOM_OUTSIDE:
            position = DependantPosition.get_random_outside().get_manim_position_center((vw, vh))

        return position
    
# TODO: These 2 methods below are similar to the ones in
# yta_multimedia\video\edition\effect\moviepy\position\objects\coordinate.py
def is_size_valid(size: tuple):
    """
    Check if the provided 'size' is a valid value or not.
    """
    if not PythonValidator.is_tuple(size) or len(size) != 2 or not NumberValidator.is_number_between(size[0], 1, NORMALIZATION_MAX_VALUE) or not NumberValidator.is_number_between(size[1], 1, NORMALIZATION_MAX_VALUE):
        return False
    
    return True

def validate_size(size: tuple, parameter_name: str):
    """
    Validate the provided 'size' and raises an Exception if
    not valid.
    """
    if not is_size_valid(size):
        raise Exception(f'The provided {parameter_name} parameter is not a tuple or does not have 2 elements that are numbers between 1 and {str(NORMALIZATION_MAX_VALUE)}.')


from yta_multimedia.video.edition.effect.m_effect import MEffect as Effect
from yta_multimedia.video.parser import VideoParser
from moviepy.video.fx import TimeMirror
from moviepy import Clip


class ReversedEffect(Effect):
    """
    This method creates a new one but in reversa, also
    with the sound reversed.
    """
    def apply(self, video: Clip) -> Clip:
        # TODO: This is not working properly yet
        #PythonValidator.validate_method_params(BlinkEffect.apply, locals(), ['video'])
        video = VideoParser.to_moviepy(video)

        return TimeMirror().apply(video)
    
    # TODO: Here below is the old method that I will keep 
    # a couple of commits before I confirm it is working
    # @classmethod
    # def apply(cls, video: Union[str, VideoFileClip, VideoClip, CompositeVideoClip, ImageClip, ColorClip]):
    #     """
    #     Applies the effect to the provided 'video'.
    #     """
    #     video = VideoEffect.parse_moviepy_video(video)

    #     reversed_frames_array = VideoFrameExtractor.get_all_frames(video)[::-1]

    #     # TODO: Try to do this audio processing in memory
    #     AUDIO_FILE = create_temp_filename('tmp_audio.mp3')
    #     REVERSED_AUDIO_FILE = create_temp_filename('tmp_reversed_audio.mp3')
    #     video.audio.write_audiofile(AUDIO_FILE, fps = 44100)
    #     AudioSegment.from_mp3(AUDIO_FILE).reverse().export(REVERSED_AUDIO_FILE)
    #     reversed_audio = AudioFileClip(REVERSED_AUDIO_FILE)

    #     return ImageSequenceClip(reversed_frames_array, fps = video.fps).with_audio(reversed_audio)


from scamp import Session, playback_settings, Envelope, wait
from random import random, choice
from enum import Enum


class Instrument(Enum):
    PIANO = 'piano'
    CELLO = 'cello'
    CLARINET = 'clarinet'
    OBOE = 'oboe'
    GUITAR = 'guitar'
    VIOLIN = 'violin'

# Inspiration: https://inspiredacoustics.com/en/MIDI_note_numbers_and_center_frequencies
class Note(Enum):
    """
    Pitches, from letters to numbers, to make easier 
    working with this MIDI notes.
    """
    A3 = 57
    B3 = 59
    C4 = 60
    D4 = 62
    E4 = 64
    F4 = 65
    G4 = 67
    A4 = 69
    B4 = 71
    C5 = 72

class InstrumentPlayer:
    session = None
    session_tempo = 60
    __parts = {}

    def __init__(self, session_tempo: int = 60):
        if not session_tempo:
            session_tempo = 60

        # TODO: Check more stuff maybe

        self.session_tempo = session_tempo

    def __start_session(self):
        if not self.session:
            self.session = Session(self.session_tempo)

        return self.session
    
    def __get_instrument_part(self, instrument: Instrument):
        """
        Gets the existing session instrument part or creates it
        and then returns it.
        """
        if not instrument:
            raise Exception('No "instrument" provided.')
        
        if not self.session:
            self.__start_session()

        if not instrument.value in self.__parts:
            self.__parts[instrument.value] = self.session.new_part(instrument.value)

        return self.__parts[instrument.value]
    
    def play_note(self, instrument: Instrument, note: Note, seconds: float = 1):
        """
        Playes the provided 'note' with the also provided 'instrument'
        during the given 'seconds'.
        """
        # TODO: By now I'm not strict with 'note' being a Note because
        # I don't have all notes registered so I could pass some int
        # values as well (by now)
        instrument_part = self.__get_instrument_part(instrument)
        instrument_part.play_note(note.value, 1, seconds)

    def play_notes(self, instrument: Instrument, notes: list[Note]):
        for note in notes:
            self.play_note(instrument, note)

    # session.start_transcribing()
    # session.stop_transcribing()


def __play_instrument_randomly(instrument: Instrument):
    #playback_settings.recording_file_path = "chello.wav"

    session = Session()
    session.tempo = 60 

    instrument = session.new_part(instrument.value)

    pitches = [60, 62, 64, 60, 64, 65, 67, 64, 67, 69, 72, 64, 72, 60] # "Frère Jacques"
    pitches = [67, 67, 69, 69, 71, 71, 69, 67, 65, 65, 67, 65, 67, 69, 67] # "Ode to Joy" de Beethoven
    pitches = [60, 62, 65, 62, 60, 62, 67, 65, 64, 62, 60] * 4
    pitches = [67, 67, 69, 71, 69, 67, 65, 64, 62, 62, 64, 65, 67, 69, 71, 69, 67] * 4

    for pitch in pitches:
        instrument.play_note(pitch, 1, 0.05)

def __test():
    """
    This method is so experimental. It uses a library that allows
    creating instrumental part of a song by using instruments and
    telling the songs you want to listen.

    TODO: This need a lot of work, investigation and refactor
    """
    # Thanks to:
    # https://www.youtube.com/watch?v=vpv686Rasds

    # This is for saving the generated audio, that is based on:
    # https://scampsters.marcevanstein.com/t/can-we-send-scamp-output-to-a-wav-file/314
    playback_settings.recording_file_path = "chello.wav"

    session = Session()
    session.tempo = 60 

    cello = session.new_part('cello')

    #session.start_transcribing()

    forte_piano = Envelope.from_levels_and_durations(
        [0.8, 0.4, 1.0], [0.2, 0.8], curve_shapes = [0, 3]
    )

    diminuendo = Envelope.from_levels([0.8, 0.3])

    def wrap_in_range(value, low, high):
        return (value - low) % (high - low) + low
    
    for pitch in (48, 53, 64, 75, 80):
        cello.play_note(pitch, 1.0, 1.0)

    interval = 1
    cello_pitch = 48

    do_continue = True
    while do_continue:
        if random() < 0.7:
            cello.play_note(cello_pitch, forte_piano, choice([1.0, 1.5]))
        else:
            cello.play_note(cello_pitch, diminuendo, choice([2.0, 2.5, 3.0]))
            wait(choice([1.0, 1.5]))
        cello_pitch = wrap_in_range(cello_pitch + interval, 36, 60)
        interval += 1

        if interval == 20:
            do_continue = False

    # ImportError: abjad was not found; LilyPond output is not available.
    #session.stop_transcribing().to_score(time_signature = '3/8').show()
    #session.stop_transcribing()

    return


from yta_multimedia.video.edition.effect.m_effect import MEffect as Effect
from yta_multimedia.video.parser import VideoParser
from yta_multimedia.video.edition.effect.moviepy.mask import ClipGenerator
from yta_general_utils.random import randrangefloat, random_bool
from moviepy import concatenate_videoclips, CompositeVideoClip, ColorClip, ImageClip
from moviepy.Clip import Clip


class DynamicPostcardEffect(Effect):
    """
    Makes the provided clip being displayed as a dynamic 
    postcard whose frames are rotated and resized to make
    an effect similar to the one in this reel:

    https://www.instagram.com/reel/C-ZzozroX_1/?igsh=MTVrcmtsYWFuMnowZw%3D%3D

    This effect is recommended to be used with a pure 
    white background.
    """
    def apply(self, video: Clip) -> Clip:
        background_video = ClipGenerator.get_default_background_video(duration = 1 / video.fps)

        return self.apply_over_video(video, background_video)
    
    def apply_over_video(self, video: Clip, background_video: Clip):
        # TODO: This is not working properly yet
        #PythonValidator.validate_method_params(BlinkEffect.apply, locals(), ['video'])
        video = VideoParser.to_moviepy(video)

        # TODO: We can use images as backgrounds if we have them
        # but we will have, for sure, dynamic white backgrounds
        # we create by ourselves
        #backgrounds_abspath = 'C:/Users/dania/Desktop/tmp_greenscreens/test_effect/'
        #files = FileHandler.get_list(backgrounds_abspath, FileSearchOption.FILES_ONLY)

        # I could obtain images from somewhere and resize all of them
        # to 1920x1080 keeping the aspect ratio

        # We need to have images enough for each video frame
        #images = files * (video.n_frames // len(files))

        frames = []
        for index, frame in enumerate(video.iter_frames()):
            # TODO: We can apply a stop-motion effect but that makes
            # the video last more than the original video
            frames.append(CompositeVideoClip([
                # TODO: What if 1920x1080 is not the original video size (?)
                ColorClip((1920, 1080), color = [255, 255, 255], duration = 1 / video.fps),
                # TODO: We can use the images as background instead of
                # these white backgrounds
                # ImageClip(images[index], duration = 5 / 60).with_fps(60),
                ImageClip(frame, duration = 1 / video.fps).with_fps(video.fps).with_position(('center', 'center'))
            ]))

        clips = []
        for frame in frames:
            #image_frame = ImageClip(frame, duration = 1 / 60).with_fps(60)
            resize = 1 + randrangefloat(0, 0.1, step = 0.005)
            rotation = 2 * randrangefloat(0, 1, step = 0.1)
            if random_bool():
                rotation = -abs(rotation)

            clips.append(CompositeVideoClip([
                ClipGenerator.get_default_background_video(duration = 1 / video.fps),
                #image_frame.with_position(('center', 'center')).resized(1 + randrangefloat(0, 0.1, step = 0.005)).rotated(rotation)
                frame.with_position(('center', 'center')).resized(resize).rotated(rotation)
            ]))

        # TODO: I think I am creating more clips than needed but,
        # as it seems to be working, I don't touch anything by now
        video = concatenate_videoclips(clips)
        # TODO: Do this as EaseIn or similar, not linear
        #white_background = ClipGenerator.generate_color_background((1920, 1080), [255, 255, 255], video.duration, video.fps)
        video = CompositeVideoClip([
            background_video,
            #ClipGenerator.get_default_background_video(duration = video.duration),
            video.resized(lambda t: 1.5 - 0.5 * t / video.duration).with_position(('center', 'center'))
        ])

        return video


from yta_general_utils.web.scrapper.chrome_scrapper import ChromeScrapper
from yta_general_utils.temp import create_temp_filename
from PIL import Image

def download_youtube_video_comments(url: str, comments_number: int = 5):
    """
    This method downloads the 'comments_number' first comments (if
    available) of Youtube video of the provided 'url'. It will 
    scrap the web page and will screenshot those comments. The 
    amount of available comments in that video could be lower than
    the requested 'comments_number' or even zero.

    This method will return an array containing the images opened.
    If there are no comments in the provided video, the array will
    be empty.
    """
    if not url:
        raise Exception('No "url" provided.')
    
    # TODO: Check if url is a valid Youtube url

    if not comments_number:
        comments_number = 5

    scrapper = ChromeScrapper(False)
    # Example of url: 'https://www.youtube.com/watch?v=OvUj2WsADjI'
    scrapper.go_to_web_and_wait_util_loaded(url)
    # We need to scroll down to let the comments load
    # TODO: This can be better, think about a more specific strategy
    # about scrolling
    scrapper.scroll_down(1000)
    scrapper.wait(1)
    scrapper.scroll_down(1000)
    scrapper.wait(1)

    # We need to make sure the comments are load
    scrapper.find_element_by_element_type_waiting('ytd-comment-thread-renderer')
    comments = scrapper.find_elements_by_element_type('ytd-comment-thread-renderer')

    if len(comments) >= comments_number:
        comments[:5]

    # We remove the header bar to avoid being over our comments in some cases
    youtube_top_bar = scrapper.find_element_by_id_waiting('masthead-container')
    scrapper.remove_element(youtube_top_bar)

    screenshots = []
    for comment in comments:
        # TODO: I need to close the 'No, gracias' 'Probar 1 mes' popup
        scrapper.scroll_to_element(comment)
        style = 'width: 500px; padding: 10px;'
        scrapper.set_element_style(comment, style)
        filename = create_temp_filename('tmp_comment_screenshot.png')
        scrapper.screenshot_element(comment, filename)
        screenshots.append(Image.open(filename))

    # TODO: Maybe I need to return the filenames instead (?)
    return screenshots


"""
    When I get the 'innerText' attribute from a comment,
    I receive it with a specific structure:

    print(comment.get_attribute('innerText'))
    # I can handle the information from this innertext
    # If I split by \n:

    # 1st is author (@pabloarielcorderovillacort2149)
    # 2nd is date (hace 6 meses)
    # 3rd and next ones are comment text
    # Penultimate is the number of likes (number)
    # Last one is 'Responder'
    #
    # This below is an example of a read comment:
    #
    # @pabloarielcorderovillacort2149
    # hace 6 meses
    # No puedo esperar tu siguiente vídeo.

    # Andy Serkis es una joya de actor, Gollum y Cesar son mis personajes favoritos de este actor.
    # 3
    # Responder
"""


from yta_multimedia.image.edition.resize import resize_image
from yta_multimedia.video.parser import VideoParser
from yta_multimedia.video.edition.resize import resize_video
from yta_general_utils.region import Region as BaseRegion
from yta_general_utils.image.parser import ImageParser


class ImageRegion(BaseRegion):
    """
    Class to represent a region built by two coordinates, one in
    the top left corner and another one in the bottom right 
    corner. This is useful to place an image inside the region.
    """
    def resize_image_to_fit_in(self, image):
        """
        This method rescales the provided 'image' to make it fit in
        this region. Once it's been rescaled, this image should be
        placed in the center of the region.
        """
        image = ImageParser.to_pillow(image)

        image = resize_image(image, (self.width, self.height))

        # We enlarge it by a 1% to avoid some balck pixels lines
        image = image.resize((image.size[0] * 1.01, image.size[1] * 1.01))

        return image
    
    # TODO: This could be private maybe
    # TODO: I should do this in yta_multimedia as it is
    # video related
    def resize_video_to_fit_in(self, video):
        """
        This method rescales the provided 'video' to make it fit in
        this region. Once it's been rescaled, this video should be
        placed in the center of the region.
        """
        video = VideoParser.to_moviepy(video)

        video = resize_video(video, (self.width, self.height))

        # We enlarge it by a 1% to avoid some black pixels lines
        video = video.resized(1.01)

        return video
    
    def place_video_inside(self, video):
        """
        This method rescales the provided 'video' to make it fit in
        this region. Once it's been rescaled, this videos is 
        positioned in the required position to fit the region.
        """
        video = self.resize_video_to_fit_in(video)

        x = (self.bottom_right.x + self.top_left.x) / 2 - video.w / 2
        y = (self.bottom_right.y + self.top_left.y) / 2 - video.h / 2

        # TODO: What about upper limits (out of bottom left bounds) (?)
        if x < 0:
            x = 0
        if y < 0:
            y = 0

        video = video.with_position((x, y))

        return video


from yta_multimedia.image.region.image_region import ImageRegion
from yta_general_utils.image.color.picker import PixelFilterFunction
from yta_general_utils.image.parser import ImageParser
from typing import Union
from PIL import Image

import numpy as np
import cv2


MIN_REGION_SIZE = (1920 / 20, 1080 / 20)
"""
The minimum size a region must have to be accepted
as a valid region.
"""

class ImageRegionFinder:
    directions = [(-1, 0), (1, 0), (0, -1), (0, 1), (-1, -1), (-1, 1), (1, -1), (1, 1)]

    @classmethod
    def is_valid(cls, x, y, image, visited, filter_func: callable):
        """
        This method verifies if the pixel is between the limits
        and is transparent and unvisited.
        """
        rows, cols, _ = image.shape

        return (0 <= x < rows and 0 <= y < cols and not visited[x, y] and filter_func(image[x, y]))

    @classmethod
    def dfs(cls, image: np.ndarray, visited, x, y, region, filter_func: callable):
        """
        A Deep First Search algorithm applied to the image to 
        obtain all the pixels connected in a region.
        """
        if not isinstance(image, np.ndarray):
            raise Exception('The provided "image" parameter is not a valid np.ndarray.')

        stack = [(x, y)]
        visited[x, y] = True
        region.append((x, y))
        
        while stack:
            cx, cy = stack.pop()
            for dx, dy in cls.directions:
                nx, ny = cx + dx, cy + dy
                if cls.is_valid(nx, ny, image, visited, filter_func):
                    visited[nx, ny] = True
                    region.append((nx, ny))
                    stack.append((nx, ny))

    @classmethod
    def is_inside(cls, small_bounds, large_bounds):
        """
        This method verifies if the bounds of a found region are
        inside another bounds to discard the smaller regions.
        """
        min_x_small, max_x_small, min_y_small, max_y_small = small_bounds
        min_x_large, max_x_large, min_y_large, max_y_large = large_bounds
        
        return (
            min_x_small >= min_x_large and max_x_small <= max_x_large and
            min_y_small >= min_y_large and max_y_small <= max_y_large
        )

    @classmethod
    def find_regions(cls, image: np.ndarray, filter_func: PixelFilterFunction) -> list[ImageRegion]:
        """
        This method looks for all the existing regions of transparent
        pixels that are connected ones to the others (neighbours). The
        'filter_func' parameter is the one that will classify the pixels
        as, for example, transparent or green. That 'filter_func' must
        be a method contained in the PixelFilterFunction class.

        This method returns the found regions as objects with 'top_left'
        and 'bottom_right' fields that are arrays of [x, y] positions
        corresponding to the corners of the found regions.
        """
        if not isinstance(image, np.ndarray):
            raise Exception('The provided "image" parameter is not a valid np.ndarray.')

        rows, cols, _ = image.shape
        visited = np.zeros((rows, cols), dtype=bool)
        regions = []
        
        for row in range(rows):
            for col in range(cols):
                # If we find a transparent pixel, we search
                if filter_func(image[row, col]) and not visited[row, col]:
                    region = []
                    cls.dfs(image, visited, row, col, region, filter_func)

                    for i, p in enumerate(region):
                        if i == 10:
                            break
                    
                    if region:
                        min_x = min(px[0] for px in region)
                        max_x = max(px[0] for px in region)
                        min_y = min(px[1] for px in region)
                        max_y = max(px[1] for px in region)

                        # These are the limits of the region
                        bounds = (min_x, max_x, min_y, max_y)
                        
                        # We need to avoid small regions contained in others and
                        # also use a minimum size to accept a region as valid
                        if max_x - min_x >= MIN_REGION_SIZE[0] and max_y - min_y >= MIN_REGION_SIZE[1] and not any(cls.is_inside(bounds, r['bounds']) for r in regions):
                            regions.append({
                                'bounds': bounds
                            })

        # I want another format, so:
        for index, region in enumerate(regions):
            regions[index] = ImageRegion(region['bounds'][2], region['bounds'][0], region['bounds'][3], region['bounds'][1])

        return regions
    
    @classmethod
    def find_green_regions(cls, image: Union[str, Image.Image, np.ndarray]) -> list[ImageRegion]:
        """
        This method returns the found green regions as objects with
        'top_left' and 'bottom_right' fields that are arrays of [x, y] 
        positions corresponding to the corners of the found regions.
        """
        image = ImageParser.to_numpy(image, 'RGB')
            
        return cls.find_regions(image, PixelFilterFunction.is_green)
    
    @classmethod
    def find_transparent_regions(cls, image: Union[str, Image.Image, np.ndarray]) -> list[ImageRegion]:
        """
        This method returns the found transparent regions as objects
        with 'top_left' and 'bottom_right' fields that are arrays of
        [x, y] positions corresponding to the corners of the found
        regions.
        """
        image = ImageParser.to_numpy(image, 'RGBA')
            
        return cls.find_regions(image, PixelFilterFunction.is_transparent)









# TODO: This below need work, is using a HSV mask to
# recognize regions, but this is actually a new class
# to bring functionality from 'yta_general_utils'

class RegionFinderTest:
    @staticmethod
    def detect_regions(image, low_range, high_range):
        # Convertir la imagen de BGR a HSV
        imagen_hsv = cv2.cvtColor(image, cv2.COLOR_BGR2HSV)
        
        # Crear una máscara con los píxeles que están dentro del rango de color
        mascara = cv2.inRange(imagen_hsv, low_range, high_range)
        
        # Encontrar los contornos de las regiones detectadas
        contornos, _ = cv2.findContours(mascara, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
        
        # Dibujar los contornos sobre la imagen original
        imagen_contornos = image.copy()
        for contorno in contornos:
            if cv2.contourArea(contorno) > 100:  # Filtrar contornos pequeños
                cv2.drawContours(imagen_contornos, [contorno], -1, (0, 255, 0), 2)
        
        return imagen_contornos, mascara

    @staticmethod
    def test():
        imagen = cv2.imread('imagen.jpg')

        # Definir el rango de color en HSV (por ejemplo, un verde brillante)
        rango_bajo = np.array([35, 50, 50])   # El valor mínimo del verde en HSV
        rango_alto = np.array([85, 255, 255])  # El valor máximo del verde en HSV

        # Llamar a la función para detectar las regiones del color
        imagen_contornos, mascara = RegionFinderTest.detect_regions(imagen, rango_bajo, rango_alto)

        # Mostrar los resultados
        cv2.imshow('Regiones Detectadas', imagen_contornos)
        cv2.imshow('Máscara', mascara)

        # Esperar a que se presione una tecla para cerrar las ventanas
        cv2.waitKey(0)
        cv2.destroyAllWindows()




from yta_multimedia.video.parser import VideoParser
from yta_multimedia.video.frames.video_frame import VideoFrame
from yta_multimedia.video.edition.duration import set_video_duration
from yta_general_utils.programming.parameter_validator import NumberValidator
from yta_multimedia.video.edition.effect.moviepy.mask import ClipGenerator
from yta_general_utils.programming.parameter_validator import PythonValidator
from typing import Union
from moviepy import VideoClip, CompositeVideoClip, concatenate_videoclips
from moviepy.Clip import Clip
from copy import copy

import numpy as np


class VideoHandler:
    """
    Class created to simplify and encapsulate some basic
    calculations related to a video, its frames and more
    properties.
    """
    video: Clip = None
    """
    The video to handle.
    """
    _frame_duration: float = None
    _frame_time_moments: list = None

    def __init__(self, video: Clip):
        video = VideoParser.to_moviepy(video)

        self.video = video

    @property
    def frame_duration(self):
        """
        The frame duration of the video, calculated with this
        formula:

        frame_duration = duration / (duration * fps)
        """
        if self._frame_duration is None:
            self._frame_duration = 1 / self.video.fps
            #self._frame_duration = self.video.duration / (self.video.duration * self.video.fps)

        return self._frame_duration
    
    @property
    def frame_time_moments(self):
        """
        The list of frame time moments (ts) to build the clip
        frame by frame. This can be used to precalculate video
        positions, resizes or rotations, in order to apply them
        later one by one, frame by frame, reading directly form
        an array.

        You can access to an specific time moment by using the
        'get_frame_time_moment' method.
        """
        if self._frame_time_moments is None:
            self._frame_time_moments = np.linspace(0, self.video.duration, self.frames_number)

        return self._frame_time_moments
    
    @property
    def frames_number(self):
        return self.video.n_frames
    
    def get_frame_number_by_time_moment(self, t: float):
        """
        Return the frame number (from 0 to the last one)
        according to the frame time moment 't' provided.

        The frame number is calculated with the next formula:

        rotations[int((t + frame_duration * 0.1) // frame_duration)]

        The '+ frame_duration * 0.1' part is needed to make sure
        each frame access is different (due to floating points
        errors)
        """
        if not NumberValidator.is_positive_number(t):
            raise Exception(f'The provided "t" parameter "{str(t)}" is not a valid frame time moment.')
        
        return VideoHandler.t_to_frame_index(t, self.frame_duration)
        # TODO: What about looking for in in the array (?)
        # Maybe I don't find it because of floating points (?)

    # TODO: Maybe move this method to a VideoMaskInverter
    # class...
    def invert(self):
        """
        Invert the received 'video' (that must be a moviepy 
        mask or normal clip) and return it inverted as a
        VideoClip. If the provided 'video' is a mask, this 
        will be also a mask.

        If the 'clip' provided is a mask clip, remember to
        set it as the new mask of your main clip.

        This inversion is a process in which the numpy array
        values of each frame are inverted by substracting the
        highest value. If the frame is an RGB frame with 
        values between 0 and 255, it will be inverted by 
        doing 255 - X on each frame pixel value. If it is
        normalized and values are between 0 and 1 (it is a 
        mask clip frame), by doing 1 - X on each mask frame
        pixel value.
        """
        mask_frames = [VideoFrame(frame).inverted() for frame in self.video.iter_frames()]

        # TODO: Which calculation is better, t * fps or t // frame_duration (?)
        return VideoClip(lambda t: mask_frames[int(t * self.video.fps)], is_mask = self.video.is_mask).with_fps(self.video.fps)
    
    def prepare_background_clip(self, background_video: Union[str, Clip]):
        """
        Prepares the provided 'background_video' by modifying its duration to
        be the same as the provided 'video'. By default, the strategy is 
        looping the 'background_video' if the 'video' duration is longer, or
        cropping it if it is shorter. This method returns the background_clip
        modified according to the provided 'video'.

        This method will raise an Exception if the provided 'video' or the provided
        'background_video' are not valid videos.

        TODO: Add a parameter to be able to customize the extend or enshort
        background strategy.
        """
        background_video = VideoParser.to_moviepy(background_video)

        background_video = set_video_duration(background_video, self.video.duration)

        return background_video

    @staticmethod
    def t_to_frame_index(t: float, frame_duration: float):
        """
        Transform the provided 't', applying the also provided
        'frame_duration' to the frame times array index to be
        able to access to its value.

        This method applies the next formula:

        int((t + frame_duration * 0.1) // frame_duration)
        """
        # This is apparently the same as int(t * self.video.fps)
        return int(t // frame_duration)
        return int((t + frame_duration * 0.1) // frame_duration)

    @staticmethod
    def concatenate_videos(videos: list[Clip]):
        """
        Concatenate the provided 'videos' but fixing the
        videos dimensions. It will wrap any video that
        doesn't fit the 1920x1080 scene size with a full
        transparent background to fit those dimensions.
        """
        if not PythonValidator.is_list(videos):
            videos = [videos]

        videos = [VideoParser.to_moviepy(video) for video in videos]
        videos = [VideoHandler(video).wrap_with_transparent_background() for video in videos]
        
        return concatenate_videoclips(videos)
    
    def wrap_with_transparent_background(self):
        """
        Put a full transparent background behind the video
        if its size is not our default scene (1920x1080) and
        places the video on the center of this background.

        This method works with a copy of the original video so
        only the returned one is changed.
        """
        # TODO: This should be a constant
        MOVIEPY_DEFAULT_SCENE_SIZE = (1920, 1080)

        video = copy(self.video)
        # TODO: Is this changing the variable value or do I
        # need to do by position (?)
        if self.video.size != MOVIEPY_DEFAULT_SCENE_SIZE:
            # I place the video at the center of the new background but
            # I reposition it to place with its center in the same place
            # as the original one
            original_center_positions = []
            # TODO: Careful with this t
            for t in self.frame_time_moments:
                pos = video.pos(t)
                original_center_positions.append((pos[0], pos[1]))

            video = CompositeVideoClip([
                ClipGenerator.get_default_background_video(duration = video.duration),
                video.with_position(('center', 'center'))
            ]).with_position(lambda t: (
                original_center_positions[VideoHandler.t_to_frame_index(t, 1 / video.fps)][0] - MOVIEPY_DEFAULT_SCENE_SIZE[0] / 2,
                original_center_positions[VideoHandler.t_to_frame_index(t, 1 / video.fps)][1] - MOVIEPY_DEFAULT_SCENE_SIZE[1] / 2
            ))

        return video


GOOGLE_SEARCH_IMAGE_GOOGLE_DRIVE_DOWNLOAD_URL = 'https://drive.google.com/file/d/1rcowE61X8c832ynh0xOt60TH1rJfcJ6z/view?usp=sharing'
YOUTUBE_SEARCH_IMAGE_GOOGLE_DRIVE_DOWNLOAD_URL = 'https://drive.google.com/file/d/11WF7Yjep1IOYPU9TAI0x5zxYMLQWMKyg/view?usp=sharing'


from yta_multimedia.utils.resize import get_cropping_points_to_keep_aspect_ratio
from yta_general_utils.image.parser import ImageParser
from yta_general_utils.programming.parameter_validator import PythonValidator
from yta_general_utils.image.converter import ImageConverter
from yta_general_utils.programming.output import handle_output_filename
from PIL import Image
from typing import Union, Any

import numpy as np
import cv2


# TODO: Refactor and encapsulate in a class
def resize_image(image: Union[str, Any], size, output_filename: Union[str, None] = None):
    """
    Resizes the image to the provided 'size' by cropping a
    region of the given 'image' that fits the 'size' aspect
    ratio and resizing that region to the 'size'.

    This method is using the whole image and then resizing,
    so the quality of the image is preserved and no small
    regions are used. The most part of the image is 
    preserved.

    This method returns the image modified.

    This method will write the image if 'output_filename' is
    provided.
    """
    image = ImageParser.to_pillow(image)

    top_left, bottom_right = get_cropping_points_to_keep_aspect_ratio(image.size, size)
    image = image.crop((top_left[0], top_left[1], bottom_right[0], bottom_right[1]))
    image = image.resize(size)

    output_filename = handle_output_filename(output_filename, 'png')
    if output_filename:
        image.save(output_filename)

    return image

# TODO: Are these methods below interesting or even used (?) I think
# they can be done with libraries .resize methods...
# TODO: Make 'image: Union[str, Image.Image, np.ndarray]' work please
def resize_image_scaling(image: Union[str, Any], width: int, height: int, output_filename: Union[str, None] = None):
    """
    Resizes the provided 'image' (if valid) to the provided dimensions
    ('width', 'height') scaling it (losing not the aspect ratio). The
    image will be stored locally as 'output_filename' if provided and
    will be returned once it's been scalled, in any case.

    TODO: Please, test and check this method
    """
    # TODO: This method needs to be checked properly
    if not image:
        raise Exception('No "image" provided.')

    if not width:
        raise Exception('No "width" provided.')
    
    if not height:
        raise Exception('No "height" provided.')

    if PythonValidator.is_string(image):
        # TODO: Check if image is actually an image
        image = Image.open(image)

    if PythonValidator.is_numpy_array(image):
        image = ImageConverter.numpy_image_to_pil(image)

    # Lets resize it
    image_width, image_height = image.size

    if image_width == width and image_height == height:
        return image.save(output_filename)

    aspect_ratio = image_width / image_height
    if aspect_ratio > (width / height):
        # Image is very horizontal, so width changes faster, we need to focus on height
        factor = (height * 100 / image_height) / 100
        image_width = int(image_width * factor)
        image_height = height
    else:
        # Image is very vertical, so height changes faster, we need to focus on width
        factor = (width * 100 / image_width) / 100
        image_width = 1920
        image_height = int(image_height * factor)
    image = image.resize((image_width, image_height))

    # We will crop form the center to edges
    left = 0
    right = width
    if image_width > width:
        # If it is 1960 => leave [0, 20], get [20, 1940], leave [1940, 1960]
        margin = int((image_width - width) / 2)
        left = 0 + margin
        right = image_width - margin
        # We make and adjustment if some pixel left
        while (right - left) > width:
            right -= 1
        while (right - left) < width:
            if left > 0:
                left -= 1
            else:
                right += 1

    top = 0
    bottom = height
    if image_height > height:
        # If it is 1140 => leave [0, 30], get [30, 1110], leave [1110, 1140]
        margin = int((image_height - height) / 2)
        top = 0 + margin
        bottom = image_height - margin
        # We make and adjustment if some pixel left
        while (bottom - top) > height:
            bottom -= 1
        while (bottom - top) < height:
            if top > 0:
                top -= 1
            else:
                bottom += 1

    # Image that is 1920x1080 and is the center of the original image
    image = image.crop((left, top, right, bottom))

    if output_filename:
        image.save(output_filename)

    return image

def resize_image_without_scaling(image: Union[str, Any], width: int, height: int, output_filename: Union[str, None] = None):
    """
    Resizes the provided 'image' (if valid) to the provided dimensions
    ('width', 'height') guaranteen not that the aspect ratio is kept. 
    The image will be stored locally as 'output_filename' if provided
    and will be returned once it's been scalled, in any case.

    TODO: Please, test and check this method
    """
    # TODO: This method needs to be checked properly
    if not image:
        raise Exception('No "image" provided.')

    if not width:
        raise Exception('No "width" provided.')
    
    if not height:
        raise Exception('No "height" provided.')

    if PythonValidator.is_string(image):
        # TODO: Check if image is actually an image
        image = Image.open(image)

    if PythonValidator.is_numpy_array(image):
        image = ImageConverter.numpy_image_to_pil(image)

    # From here: https://stackoverflow.com/a/14140796
    image = image.convert('RGB')
    image = np.array(image) # To cv2
    image = image[:, :, ::-1].copy() # Convert RGB to BGR

    #image = cv2.imread(image_filename)
    resized_image = cv2.resize(image, dsize = (width, height), interpolation = cv2.INTER_CUBIC)

    if output_filename:
        cv2.imwrite(output_filename, resized_image)

    return resized_image


# TODO: Remove these 2 methods below if the ones above are working, because
# they are more general (using different types or images)
def resize_image_file_scaling(image_filename: str, width: int, height: int, output_filename = None):
    """
    Resizes the provided 'image_filename' to the provided 'width' and 'height' keeping the
    aspect ratio. This method enlarges the image to fit the desired size and then makes a 
    crop to obtain that size from the center of the resized image. If 'output_filename' is
    provided, the image is saved locally with that name. If not, it is only returned
    """
    if not image_filename:
        return None
    
    # TODO: Do some checkings on width and height (?)

    # TODO: This is not working well when '' passed
    if output_filename is not None and not output_filename:
        return None

    image = Image.open(image_filename)
    image_width, image_height = image.size

    if image_width == width and image_height == height:
        return image.save(output_filename)

    aspect_ratio = image_width / image_height
    if aspect_ratio > (width / height):
        # Image is very horizontal, so width changes faster, we need to focus on height
        factor = (height * 100 / image_height) / 100
        image_width = int(image_width * factor)
        image_height = height
    else:
        # Image is very vertical, so height changes faster, we need to focus on width
        factor = (width * 100 / image_width) / 100
        image_width = 1920
        image_height = int(image_height * factor)
    image = image.resize((image_width, image_height))

    # We will crop form the center to edges
    left = 0
    right = width
    top = 0
    bottom = height
    if image_width > width:
        # If it is 1960 => leave [0, 20], get [20, 1940], leave [1940, 1960]
        margin = int((image_width - width) / 2)
        left = 0 + margin
        right = image_width - margin
        # We make and adjustment if some pixel left
        while (right - left) > width:
            right -= 1
        while (right - left) < width:
            if left > 0:
                left -= 1
            else:
                right += 1
    if image_height > height:
        # If it is 1140 => leave [0, 30], get [30, 1110], leave [1110, 1140]
        margin = int((image_height - height) / 2)
        top = 0 + margin
        bottom = image_height - margin
        # We make and adjustment if some pixel left
        while (bottom - top) > height:
            bottom -= 1
        while (bottom - top) < height:
            if top > 0:
                top -= 1
            else:
                bottom += 1

    image = image.crop((left, top, right, bottom))
    # Image that is 1920x1080 and is the center of the original image
    if output_filename:
        image.save(output_filename)

    return image